import unittest
import logging
import sys
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class HuggingFaceCallbackTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(Uploader, 'upload_signal')
    def test_callback(self, mocked_upload_signal):
        import torch
        if not torch.cuda.is_available():
            return
        from datasets import load_dataset
        raw_datasets = load_dataset("imdb")

        from transformers import AutoTokenizer
        tokenizer = AutoTokenizer.from_pretrained("bert-base-cased", cache_dir='temp')

        def tokenize_function(examples):
            return tokenizer(examples["text"],
                             padding="max_length", truncation=True)

        tokenized_datasets = raw_datasets.map(tokenize_function, batched=True)

        small_train_dataset = tokenized_datasets["train"].shuffle(
            seed=42).select(
            range(10))
        small_eval_dataset = tokenized_datasets["test"].shuffle(
            seed=42).select(
            range(10))
        full_train_dataset = tokenized_datasets["train"]
        full_eval_dataset = tokenized_datasets["test"]

        from transformers import AutoModelForSequenceClassification
        model = AutoModelForSequenceClassification.from_pretrained(
            "distilbert-base-cased", cache_dir='temp', num_labels=2)

        from transformers import TrainingArguments
        training_args = TrainingArguments("test_trainer", num_train_epochs=1)
            
        from transformers import Trainer
        from graphsignal.callbacks.huggingface import GraphsignalPTCallback
        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=small_train_dataset,
            eval_dataset=small_eval_dataset)
        trainer.add_callback(GraphsignalPTCallback())

        trainer.train()

        signal = mocked_upload_signal.call_args[0][0]

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        self.assertEqual(signal.endpoint_name, 'training_step')

        self.assertTrue(
            signal.agent_info.framework_profiler_type, 
            signals_pb2.AgentInfo.ProfilerType.HUGGING_FACE_PROFILER)

        self.assertEqual(
            signal.frameworks[-1].type,
            signals_pb2.FrameworkInfo.FrameworkType.HUGGING_FACE_FRAMEWORK)
        self.assertTrue(signal.frameworks[-1].version.major > 0)

        self.assertTrue(len(signal.op_stats) > 0)



import unittest
import logging
import sys
import os
import json
from unittest.mock import patch, Mock
import tensorflow as tf
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.profilers.tensorflow import TensorFlowProfiler
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class TensorFlowProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_read_info(self):
        profiler = TensorFlowProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.read_info(signal)

        self.assertEqual(
            signal.frameworks[0].type,
            signals_pb2.FrameworkInfo.FrameworkType.TENSORFLOW_FRAMEWORK)

    def test_start_stop(self):
        @tf.function
        def f(x):
            while tf.reduce_sum(x) > 1:
                #tf.print(x)
                x = tf.tanh(x)
            return x

        profiler = TensorFlowProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.start(signal)
        f(tf.random.uniform([5]))
        profiler.stop(signal)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        self.assertTrue(sum([op_stats.count for op_stats in signal.op_stats]) > 0)
        if len(tf.config.list_physical_devices('GPU')) > 0:
            self.assertTrue(sum([op_stats.total_device_time_us for op_stats in signal.op_stats]) > 0)
            self.assertTrue(sum([op_stats.self_device_time_us for op_stats in signal.op_stats]) > 0)
        else:
            self.assertTrue(sum([op_stats.total_host_time_us for op_stats in signal.op_stats]) > 0)
            self.assertTrue(sum([op_stats.self_host_time_us for op_stats in signal.op_stats]) > 0)
        self.assertNotEqual(signal.trace_data, b'')





class OperationProfiler():
    def is_ready(self):
        return True

    def read_info(self, signal):
        raise NotImplementedError()

    def start(self, signal):
        raise NotImplementedError()

    def stop(self, signal):
        raise NotImplementedError()



from typing import Optional
import logging
import transformers
from transformers import TrainerCallback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver
from graphsignal.profilers.operation_profiler import OperationProfiler

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(TrainerCallback):
    def __init__(self, tags: Optional[dict] = None, profiler: Optional[OperationProfiler] = None):
        super().__init__()
        self._profiler = profiler
        self._hf_version = None
        self._tags = tags
        self._trace = None
        self._step_idx = -1

    def on_train_begin(self, args, state, control, **kwarg):
        self._configure_profiler(args)

    def on_train_end(self, args, state, control, **kwarg):
        graphsignal.upload()

    def on_step_begin(self, args, state, control, **kwarg):
        self._step_idx += 1
        self._start_trace('training_step', args, state)

    def on_step_end(self, args, state, control, **kwarg):
        self._stop_trace(args, state)

    def _start_trace(self, endpoint, args, state):
        if not self._trace:
            self._trace = graphsignal.start_trace(
                endpoint=endpoint,
                tags=self._tags,
                profiler=self._profiler)
            self._trace.set_tag('step', self._step_idx)

    def _stop_trace(self, args, state):
        if self._trace:
            if self._trace.is_tracing():
                self._update_signal(args, state)
            self._trace.stop()
            self._trace = None

    def _configure_profiler(self, args):
        try:
            self._hf_version = signals_pb2.SemVer()
            parse_semver(self._hf_version, transformers.__version__)
        except Exception:
            logger.error('Error configuring Hugging Face profiler', exc_info=True)

    def _update_signal(self, args, state):
        try:
            signal = self._trace._signal

            signal.agent_info.framework_profiler_type = signals_pb2.AgentInfo.ProfilerType.HUGGING_FACE_PROFILER

            framework = signal.frameworks.add()
            framework.type = signals_pb2.FrameworkInfo.FrameworkType.HUGGING_FACE_FRAMEWORK
            framework.version.CopyFrom(self._hf_version)
        except Exception as exc:
            logger.debug('Error in Hugging Face callback', exc_info=True)
            self._trace._add_profiler_exception(exc)


class GraphsignalPTCallback(GraphsignalCallback):
    def __init__(self, tags: Optional[dict] = None):
        from graphsignal.profilers.pytorch import PyTorchProfiler
        super().__init__(
            tags=tags,
            profiler=PyTorchProfiler())


class GraphsignalTFCallback(GraphsignalCallback):
    def __init__(self, tags: Optional[dict] = None):
        from graphsignal.profilers.tensorflow import TensorflowProfiler
        super().__init__(
            tags=tags,
            profiler=TensorflowProfiler())